from __future__ import absolute_import
from ..errs import ValidationError
from .comm import min_max

def validate_float(obj, ret, val, ctx):
    min_max(obj, ret, False)
    min_max(obj, ret, True)
    if obj.multipleOf and ret % obj.multipleOf != 0:
        raise ValidationError('{0} should be multiple of {1}'.format(val, obj.multipleOf))

    return val

def create_float(obj, v, ctx=None):
    r = float(v)
    validate_float(obj, r, v, ctx)
    return r





from __future__ import absolute_import
from ...scan import Dispatcher
from ...spec.v2_0.objects import (
    Operation
    )
import six


class YamlFixer(object):
    """ fix objects loaded by pyaml """

    class Disp(Dispatcher): pass

    @Disp.register([Operation])
    def _op(self, _, obj, app):
        """ convert status code in Responses from int to string
        """
        if obj.responses == None: return 

        tmp = {}
        for k, v in six.iteritems(obj.responses):
            if isinstance(k, six.integer_types):
                tmp[str(k)] = v
            else:
                tmp[k] = v
        obj.update_field('responses', tmp)




from __future__ import absolute_import
import uuid
import six


class UUID(object):
    """ wrapper of uuid.UUID
    """

    def __str__(self):
        return str(self.v)

    def to_json(self):
        return str(self.v)

    def apply_with(self, _, val, ctx):
        """ constructor

        :param val: things used to construct uuid
        :type val: uuid as byte, string, or uuid.UUID
        """
        if isinstance(val, uuid.UUID):
            self.v = val
        elif isinstance(val, six.string_types):
            try:
                self.v = uuid.UUID(val)
            except ValueError:
                self.v = uuid.UUID(bytes=val)
        elif isinstance(val, six.binary_type):
            # TODO: how to support bytes_le?
            self.v = uuid.UUID(bytes=val)
        else:
            raise ValueError('Unrecognized type for UUID: ' + str(type(val)))



import json
import six
from .comm import PrimJSONEncoder


class MimeCodec:
    def __init__(self):
        self._codecs = {}
        self.register('text/plain', PlainCodec())
        jsonCodec = JsonCodec()
        self.register('application/json', jsonCodec)
        #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#- manatlan's add
        self.register('application/hal+json', jsonCodec)        
        #-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-#-
        self.register('text/json', jsonCodec)

    def register(self, mime, codec):
        self._codecs[mime.lower()] = codec

    def unregister(self, mime):
        self._codecs.pop(mime.lower())

    def codec(self, mime):
        mime = mime.split(';', 1)[0]
        mime = mime.strip().lower()
        return self._codecs.get(mime, None)

    def marshal(self, mime, value, **kwargs):
        codec = self.codec(mime)
        if not codec:
            raise Exception('Could not find codec for %s, value: %s, args: %s' % (mime, value, kwargs))
        return codec.marshal(value, **kwargs)

    def unmarshal(self, mime, data, **kwargs):
        codec = self.codec(mime)
        if not codec:
            raise Exception('Could not find codec for %s, data: %s, args: %s' % (mime, data, kwargs))
        return codec.unmarshal(data, **kwargs)


class PlainCodec:
    def marshal(self, value, **kwargs):
        return value

    def unmarshal(self, data, **kwargs):
        return data


class JsonCodec:
    def marshal(self, value, **kwargs):
        return json.dumps(value, cls=PrimJSONEncoder)

    def unmarshal(self, data, **kwargs):
        if isinstance(data, six.binary_type):
            data = data.decode('utf-8')
        return json.loads(data)



# from __future__ import absolute_import
# from ...core import BaseClient
# import webapp2
# import six
# import sys


# class Webapp2TestClient(BaseClient):
#     """ Webapp2 Test Client
#     """

#     __schemes__ = set(['http', 'https'])

#     def __init__(self, app, auth=None, keep_cookie=False, **kw):
#         """ constructor

#         :param app webapp2.WSGIApplication: the WSGI application to be tested
#         :param auth appswag.Security: auth info holder
#         :param keep_cookie bool: keep cookie as session or not
#         """
#         super(Webapp2TestClient, self).__init__(auth)

#         #
#         # if webapp2 finally support python 3, please submit an issue to remove this check
#         #
#         if sys.version_info[0] > 2:
#             raise Exception('webapp2 only support python 2.x')

#         self.__app = app
#         self.__keep_cookie = keep_cookie
#         self.__cookie = None
#         self.__kw = kw

#     def request(self, req_and_resp, opt=None, headers=None):
#         """
#         """

#         # make sure all prepared state are clean before processing
#         req, resp = req_and_resp
#         req.reset()
#         resp.reset()

#         opt = opt or {}
#         req, resp = super(Webapp2TestClient, self).request((req, resp), opt)

#         req.prepare(scheme=self.prepare_schemes(req), handle_files=True)
#         req._patch(opt)

#         composed_headers = self.compose_headers(req, headers, opt)

#         url = req.url
#         if req.query:
#             if url[-1] not in ('?', '&'):
#                 url += '&' if ('?' in url) else '?'
#             url += six.moves.urllib.parse.urlencode(req.query)

#         # initiate an request every time
#         _req = webapp2.Request.blank(
#             url,
#             headers=composed_headers,
#             POST=req.data,
#             **self.__kw
#         )
#         _req.method = req.method.upper()
#         if self.__cookie:
#             _req.headers['Cookie'] = self.__cookie

#         # perform the request
#         _resp = _req.get_response(self.__app)

#         # prepare response
#         resp.apply_with(
#             status=_resp.status_int,
#             raw=_resp.body,
#             header=_resp.headerlist,
#             )

#         if self.__keep_cookie:
#             if 'Set-Cookie' in _resp.headers:
#                 self.__cookie = _resp.headers['Set-Cookie']

#         return resp




from __future__ import absolute_import
from ...core import BaseClient
from requests import Session, Request
import six


class Client(BaseClient):
    """ Client implementation based on requests
    """

    __schemes__ = set(['http', 'https'])

    def __init__(self, auth=None, send_opt=None):
        """ constructor

        :param auth appswag.SwaggerAuth: auth info used when requesting
        :param send_opt dict: options used in requests.send, ex verify=False
        """
        super(Client, self).__init__(auth)
        if send_opt is None:
            send_opt = {}

        self.__s = Session()
        self.__send_opt = send_opt

    def request(self, req_and_resp, opt=None, headers=None):
        """
        """

        # make sure all prepared state are clean before processing
        req, resp = req_and_resp
        req.reset()
        resp.reset()

        opt = opt or {}
        req, resp = super(Client, self).request((req, resp), opt)

        # apply request-related options before preparation.
        req.prepare(scheme=self.prepare_schemes(req), handle_files=False)
        req._patch(opt)

        composed_headers = self.compose_headers(req, headers, opt, as_dict=True)

        # prepare for uploaded files
        file_obj = []
        def append(name, obj):
            f = obj.data or open(obj.filename, 'rb')
            if 'Content-Type' in obj.header:
                file_obj.append((name, (obj.filename, f, obj.header['Content-Type'])))
            else:
                file_obj.append((name, (obj.filename, f)))

        for k, v in six.iteritems(req.files):
            if isinstance(v, list):
                for vv in v:
                    append(k, vv)
            else:
                append(k, v)

        rq = Request(
            method=req.method.upper(),
            url=req.url,
            params=req.query,
            data=req.data,
            headers=composed_headers,
            files=file_obj
        )
        rq = self.__s.prepare_request(rq)
        rs = self.__s.send(rq, stream=True, **self.__send_opt)

        resp.apply_with(
            status=rs.status_code,
            header=rs.headers,
            raw=six.BytesIO(rs.content).getvalue()
        )

        return resp




from __future__ import absolute_import
from ...scan import Dispatcher
from ...utils import normalize_jr
from ...spec.v2_0.objects import (
    Schema,
    Parameter,
    Response,
    PathItem,
    )


class NormalizeRef(object):
    """ normalized all $ref """

    class Disp(Dispatcher): pass

    def __init__(self, base_url):
        self.base_url = base_url

    @Disp.register([Schema, Parameter, Response, PathItem])
    def _resolve(self, path, obj, _):
        obj.update_field('$ref', normalize_jr(getattr(obj, '$ref'), self.base_url))



from __future__ import absolute_import
from ...scan import Dispatcher
from ...spec.v2_0.objects import (
    Schema
    )
import six


class Validate(object):
    """
    """
    class Disp(Dispatcher): pass

    def __init__(self):
        self.errs = []

    @Disp.result
    def result(self, result):
        """ aggregate errors """
        if result and len(result[2]) > 0:
            self.errs.extend([((result[0], result[1]), err) for err in result[2]])

    @Disp.register([Schema])
    def _validate_schema(self, path, obj, _):
        errs = []

        for v in obj.required:
            if v in obj.properties and obj.properties[v].readOnly:
                errs.append('ReadOnly property in required list: {0}'.format(v))
            # TODO: validator runs before Resolver, so we can't go through all 'allOf'
            #       Schema objects to look for the 'required' property, it's a limitation now.
            #
            #       Maybe we need to consider making another validator runs after Resolver.

        return path, obj.__class__.__name__, errs



from __future__ import absolute_import
from .primitives.comm import PrimJSONEncoder
from .utils import final, deref, CaseInsensitiveDict
from appswag import errs
from uuid import uuid4
import six
import io, codecs
import collections
import logging


logger = logging.getLogger(__name__)


class Request(object):
    """ Request layer
    """
    # TODO: make a new class for 'prepared' Request

    # option: url_netloc, replace netloc part in url, useful
    # when testing a set of Swagger APIs locally.
    opt_url_netloc = 'url_netloc'
    opt_url_scheme = 'url_scheme'

    def __init__(self, op, params):
        """ constrcutor

        :param Operation op: the related Operation object
        :param dict params: parameter set provided by user
        """

        self.__op = op
        self.__p = params
        self.__url = self.__op.url
        self.__path = self.__op.path
        self.__header = CaseInsensitiveDict()

        self.__consume = None
        self.__produce = None
        self.__scheme = None

    def reset(self):
        self.__url = self.__op.url
        self.__path = self.__op.path
        self.__header = CaseInsensitiveDict()
        self.__data = None

    def consume(self, consume):
        self.__consume = consume
        return self

    def produce(self, produce):
        self.__produce = produce
        return self

    def _prepare_forms(self):
        """ private function to prepare content for paramType=form
        """
        content_type = 'application/x-www-form-urlencoded'
        if self.__op.consumes and content_type not in self.__op.consumes:
            raise errs.SchemaError('content type {0} does not present in {1}'.format(content_type, self.__op.consumes))

        return content_type, six.moves.urllib.parse.urlencode(self.__p['formData'])

    def _prepare_body(self):
        """ private function to prepare content for paramType=body
        """
        content_type = self.__consume
        if not content_type:
            content_type = self.__op.consumes[0] if self.__op.consumes else 'application/json'
        if self.__op.consumes and content_type not in self.__op.consumes:
            raise errs.SchemaError('content type {0} does not present in {1}'.format(content_type, self.__op.consumes))

        # according to spec, payload should be one and only,
        # so we just return the first value in dict.
        for parameter in self.__op.parameters:
            parameter = final(parameter)
            if getattr(parameter, 'in') == 'body':
                schema = deref(parameter.schema)
                _type = schema.type
                _format = schema.format
                name = schema.name
                body = self.__p['body'][parameter.name]
                return content_type, self.__op._mime_codec.marshal(content_type, body, _type=_type, _format=_format, name=name)
        return None, None

    def _prepare_files(self, encoding):
        """ private function to prepare content for paramType=form with File
        """
        content_type = 'multipart/form-data'
        if self.__op.consumes and content_type not in self.__op.consumes:
            raise errs.SchemaError('content type {0} does not present in {1}'.format(content_type, self.__op.consumes))

        boundary = uuid4().hex
        content_type += '; boundary={0}'
        content_type = content_type.format(boundary)

        # init stream
        body = io.BytesIO()
        w = codecs.getwriter(encoding)

        def append(name, obj):
            body.write(six.b('--{0}\r\n'.format(boundary)))

            # header
            w(body).write('Content-Disposition: form-data; name="{0}"; filename="{1}"'.format(name, obj.filename))
            body.write(six.b('\r\n'))
            if 'Content-Type' in obj.header:
                w(body).write('Content-Type: {0}'.format(obj.header['Content-Type']))
                body.write(six.b('\r\n'))
            if 'Content-Transfer-Encoding' in obj.header:
                w(body).write('Content-Transfer-Encoding: {0}'.format(obj.header['Content-Transfer-Encoding']))
                body.write(six.b('\r\n'))
            body.write(six.b('\r\n'))

            # body
            if not obj.data:
                with open(obj.filename, 'rb') as f:
                    body.write(f.read())
            else:
                data = obj.data.read()
                if isinstance(data, six.text_type):
                    w(body).write(data)
                else:
                    body.write(data)

            body.write(six.b('\r\n'))

        for k, v in self.__p['formData']:
            body.write(six.b('--{0}\r\n'.format(boundary)))

            w(body).write('Content-Disposition: form-data; name="{0}"'.format(k))
            body.write(six.b('\r\n'))
            body.write(six.b('\r\n'))

            w(body).write(v)

            body.write(six.b('\r\n'))

        # begin of file section
        for k, v in six.iteritems(self.__p['file']):
            if isinstance(v, list):
                for vv in v:
                    append(k, vv)
            else:
                append(k, v)

        # final boundary
        body.write(six.b('--{0}--\r\n'.format(boundary)))

        return content_type, body.getvalue()

    def _patch(self, opt={}):
        """ private function to patch this request. This function
        could be called before/after preparation.

        :param dict opt: options, used options would be popped. Refer to Request.opt_* for details.
        """
        opt_netloc = opt.pop(Request.opt_url_netloc, None)
        opt_scheme = opt.pop(Request.opt_url_scheme, None)
        if opt_netloc or opt_scheme:
            scheme, netloc, path, params, query, fragment = six.moves.urllib.parse.urlparse(self.__url)
            self.__url = six.moves.urllib.parse.urlunparse((
                opt_scheme or scheme,
                opt_netloc or netloc,
                path,
                params,
                query,
                fragment
            ))

            logger.info('patching url: [{0}]'.format(self.__url))

    def prepare(self, scheme='http', handle_files=True, encoding='utf-8'):
        """ make this request ready for Clients

        :param str scheme: scheme used in this request
        :param bool handle_files: False to skip multipart/form-data encoding
        :param str encoding: encoding for body content.
        :rtype: Request
        """

        if isinstance(scheme, list):
            if self.__scheme is None:
                scheme = scheme.pop()
            else:
                if self.__scheme in scheme:
                    scheme = self.__scheme
                else:
                    raise Exception('preferred scheme:{} is not supported by the client or spec:{}'.format(self.__scheme, scheme))
        elif not isinstance(scheme, six.string_types):
            raise ValueError('"scheme" should be a list or string')

        # combine path parameters into path
        # TODO: 'dot' is allowed in swagger, but it won't work in python-format
        path_params = {}
        for k, v in six.iteritems(self.__p['path']):
            path_params[k] = six.moves.urllib.parse.quote_plus(v)

        self.__path = self.__path.format(**path_params)

        # combine path parameters into url
        self.__url = ''.join([scheme, ':', self.__url.format(**path_params)])

        # header parameters
        self.__header.update(self.__p['header'])

        # update data parameter
        content_type = None
        if self.__p['file']:
            if handle_files:
                content_type, self.__data = self._prepare_files(encoding)
            else:
                # client that can encode multipart/form-data, should
                # access form-data via data property and file from file
                # property.

                # only form data can be carried along with files,
                self.__data = self.__p['formData']

        elif self.__p['formData']:
            content_type, self.__data = self._prepare_forms()
        elif self.__p['body']:
            content_type, self.__data = self._prepare_body()
        else:
            self.__data = None

        if content_type:
            self.__header.update({'Content-Type': content_type})

        accept = self.__produce
        if not accept and self.__op.produces:
            accept = self.__op.produces[0]
        if accept:
            if self.__op.produces and accept not in self.__op.produces:
                raise errs.SchemaError('accept {0} does not present in {1}'.format(accept, self.__op.produces))
            self.__header.update({'Accept': accept})

        return self

    @property
    def scheme(self):
        """ preferred scheme used in this request
        """
        return self.__scheme

    @scheme.setter
    def scheme(self, scheme):
        self.__scheme = scheme

    @property
    def url(self):
        """ url of this request, only valid after 'prepare'

        :type: str
        """
        return self.__url

    @property
    def path(self):
        """ path of this request, only valid after 'prepare'

        :type: str
        """
        return self.__path

    @property
    def base_path(self):
        """ base path of this request

        :type: str
        """
        return self.__op.base_path

    @property
    def query(self):
        """ query part of this request

        :type: dict
        """
        return self.__p['query']

    @property
    def method(self):
        """ HTTP verb of this request

        :type: str
        """
        return self.__op.method

    @property
    def header(self):
        """ header of this request, only valid after 'prepare'

        :type: appswag.util.CaseInsensitiveDict
        """
        return self.__header

    @property
    def data(self):
        """ data carried by this request, only valid after 'prepare'

        :type: byte
        """
        return self.__data

    @property
    def files(self):
        """ files of this Request

        :type: dict of (name, primitives.File)
        """
        return self.__p['file']

    @property
    def schemes(self):
        """ required schemes for current Operation.

        :type: list of str
        """
        return self.__op.cached_schemes

    @property
    def _p(self):
        """ access internal placeholder of all parameters,
        for unittest/debug/internal purpose
        """
        return self.__p

    @property
    def _security(self):
        """ list of authorizations required

        :type: dict of list of Authorizations object.
        """
        return self.__op.security


class Response(object):
    """ Response layer
    """

    def __init__(self, op):
        """ constructor

        :param Operation op: the related Operation object
        """
        self.__op = op
        self.__raw = self.__data = None

        # init properties
        self.__status = None
        self.__header = CaseInsensitiveDict()

        # options
        self.__raw_body_only = False

    def reset(self):
        self.__status = None
        self.__header = CaseInsensitiveDict()
        self.__raw = self.__data = None
        self.__path = self.__op.path
        self.__url = self.__op.url

    def _convert_header(self, resp, k, v):
        if resp and resp.headers and k in resp.headers:
            v = resp.headers[k]._prim_(v, self.__op._prim_factory, ctx=dict(read=True))

        if k in self.__header:
            self.__header[k].append(v)
        else:
            self.__header[k] = [v]

    def apply_with(self, status=None, raw=None, header=None):
        """ update header, status code, raw datum, ...etc

        :param int status: status code
        :param str raw: body content
        :param dict header: header section
        :return: return self for chaining
        :rtype: Response
        """

        if status != None:
            self.__status = status

        r = (final(self.__op.responses.get(str(self.__status), None)) or
             final(self.__op.responses.get('default', None)))

        if header != None:
            if isinstance(header, (collections.Mapping, collections.MutableMapping)):
                for k, v in six.iteritems(header):
                    self._convert_header(r, k, v)
            else:
                for k, v in header:
                    self._convert_header(r, k, v)

        if raw != None:
            # update 'raw'
            self.__raw = raw

            if self.__status == None:
                raise Exception('Update status code before assigning raw data')

            if r and r.schema and not self.__raw_body_only:
                # update data from Opeartion if succeed else from responseMessage.responseModel
                content_type = 'application/json'
                for k, v in six.iteritems(self.header):
                    if k.lower() == 'content-type':
                        content_type = v[0].lower()
                        break
                schema = deref(r.schema)
                _type = schema.type
                _format = schema.format
                name = schema.name
                data = self.__op._mime_codec.unmarshal(content_type, self.raw, _type=_type, _format=_format, name=name)
                self.__data = r.schema._prim_(data, self.__op._prim_factory, ctx=dict(read=True))

        return self

    def raw_body_only(self, only):
        """ an option to disable parsing bytes-stream to
        appswag primitives of body response. 'True' to enable this option
        """
        self.__raw_body_only = only

    raw_body_only = property(None, raw_body_only)

    @property
    def status(self):
        """ status code

        :type: int
        """
        return self.__status

    @property
    def data(self):
        """ responsed data

        :type: primitives.Model
        """
        return self.__data

    @property
    def raw(self):
        """ raw response

        :type: str
        """
        return self.__raw

    @property
    def header(self):
        """ header of Response

        :type: dict of list, ex. {'Content-Type': [xxx, xxx]}
        """
        return self.__header


SwaggerRequest = Request
SwaggerResponse = Response



from __future__ import absolute_import
from .comm import min_max


def validate_int(obj, ret, val, ctx):
    min_max(obj, ret, False)
    min_max(obj, ret, True)

    return val
 
def create_int(obj, v, ctx=None):
    r = int(v)
    validate_int(obj, r, v, ctx)
    return r

